@@ -1,7 +1,7 @@
 
 # cfg.py - module for accessing configuration information
 #
-# Copyright (C) 2010, 2011, 2012 Arthur de Jong
+# Copyright (C) 2010, 2011, 2012, 2013 Arthur de Jong
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -20,12 +20,9 @@
 
 import logging
 import re
-import sys
 
 import ldap
 
-from expr import Expression
-
 
 # the number of threads to start
 threads = 5
@@ -33,26 +30,27 @@ threads = 5
 uid = None
 # the group id nslcd should be run as
 gid = None
+# the configured loggers
+logs = []
 
 # the LDAP server to use
-# FIXME: support multiple servers and have a fail-over mechanism
-uri = None
+uri = None  # FIXME: support multiple servers and have a fail-over mechanism
 # LDAP protocol version to use (perhaps fix at 3?)
 ldap_version = ldap.VERSION3
 # the DN to use when binding
-binddn = None
-bindpw = None
+binddn = None  # FIXME: add support
+bindpw = None  # FIXME: add support
 # the DN to use to perform password modifications as root
 rootpwmoddn = None
 rootpwmodpw = None
 
 # SASL configuration
-sasl_mech = None
-sasl_realm = None
-sasl_authcid = None
-sasl_authzid = None
-sasl_secprops = None
-sasl_canonicalize = None
+sasl_mech = None  # FIXME: add support
+sasl_realm = None  # FIXME: add support
+sasl_authcid = None  # FIXME: add support
+sasl_authzid = None  # FIXME: add support
+sasl_secprops = None  # FIXME: add support
+sasl_canonicalize = None  # FIXME: add support
 
 # LDAP bases to search
 bases = []
@@ -63,10 +61,10 @@ deref = ldap.DEREF_NEVER
 referrals = True
 
 # timing configuration
-bind_timelimit = 10
+bind_timelimit = 10  # FIXME: add support
 timelimit = ldap.NO_LIMIT
-idle_timelimit = 0
-reconnect_sleeptime = 1
+idle_timelimit = 0  # FIXME: add support
+reconnect_sleeptime = 1  # FIXME: add support
 reconnect_retrytime = 10
 
 # SSL/TLS options
@@ -79,24 +77,35 @@ tls_ciphers = None
 tls_cert = None
 tls_key = None
 
-
 # other options
-pagesize = 0
+pagesize = 0  # FIXME: add support
 nss_initgroups_ignoreusers = set()
 nss_min_uid = 0
+nss_nested_groups = False
 validnames = re.compile(r'^[a-z0-9._@$][a-z0-9._@$ \\~-]{0,98}[a-z0-9._@$~-]$', re.IGNORECASE)
 pam_authz_searches = []
 pam_password_prohibit_message = None
+reconnect_invalidate = set()
 
 
 # allowed boolean values
 _boolean_options = {'on': True, 'yes': True, 'true': True, '1': True,
                     'off': False, 'no': False, 'false': False, '0': False}
 
+# allowed log levels (we log notice which is unsupported in Python to warning)
+_log_levels = {'crit': logging.CRITICAL, 'error': logging.ERROR,
+               'err': logging.ERROR, 'warning': logging.WARNING,
+               'notice': logging.WARNING, 'info': logging.INFO,
+               'debug': logging.DEBUG, 'none': logging.INFO}
+
 # allowed values for scope option
+if not hasattr(ldap, 'SCOPE_CHILDREN') and ldap.VENDOR_VERSION >= 20400:
+    ldap.SCOPE_CHILDREN = 3   # OpenLDAP extension
 _scope_options = dict(sub=ldap.SCOPE_SUBTREE, subtree=ldap.SCOPE_SUBTREE,
                       one=ldap.SCOPE_ONELEVEL, onelevel=ldap.SCOPE_ONELEVEL,
                       base=ldap.SCOPE_BASE)
+if hasattr(ldap, 'SCOPE_CHILDREN'):
+    _scope_options['children'] = ldap.SCOPE_CHILDREN
 
 # allowed values for the deref option
 _deref_options = dict(never=ldap.DEREF_NEVER,
@@ -122,6 +131,7 @@ def _get_maps():
     # separate function as not to pollute the namespace and avoid import loops
     import alias, ether, group, host, netgroup, network, passwd
     import protocol, rpc, service, shadow
+    import sys
     return dict(
             alias=alias, aliases=alias,
             ether=ether, ethers=ether,
@@ -165,7 +175,7 @@ def read(filename):
             globals()[m.group('keyword').lower()] = int(m.group('value'))
             continue
         # parse options with a single boolean argument
-        m = re.match('(?P<keyword>referrals)\s+(?P<value>%s)' %
+        m = re.match('(?P<keyword>referrals|nss_nested_groups)\s+(?P<value>%s)' %
                          '|'.join(_boolean_options.keys()),
                      line, re.IGNORECASE)
         if m:
@@ -183,6 +193,13 @@ def read(filename):
         if m:
             globals()[m.group('keyword').lower()] = m.group('value')
             continue
+        # log <SCHEME> [<LEVEL>]
+        m = re.match('log\s+(?P<scheme>syslog|/\S*)(\s+(?P<level>%s))?' %
+                         '|'.join(_log_levels.keys()),
+                     line, re.IGNORECASE)
+        if m:
+            logs.append((m.group('scheme'), _log_levels[str(m.group('level')).lower()]))
+            continue
         # uri <URI>
         m = re.match('uri\s+(?P<uri>\S+)', line, re.IGNORECASE)
         if m:
@@ -216,7 +233,7 @@ def read(filename):
                      line, re.IGNORECASE)
         if m:
             mod = maps[str(m.group('map')).lower()]
-            mod.scope = _scope_options[m.group('keyword').lower()]
+            mod.scope = _scope_options[m.group('value').lower()]
             continue
         # map <MAP> <ATTRIBUTE> <ATTMAPPING>
         m = re.match('map\s+(?P<map>%s)\s+(?P<attribute>\S+)\s+(?P<value>\S.*)' %
@@ -255,6 +272,7 @@ def read(filename):
         # pam_authz_search <FILTER>
         m = re.match('pam_authz_search\s+(?P<value>\S.*)', line, re.IGNORECASE)
         if m:
+            from expr import Expression
             pam_authz_searches.append(Expression(m.group('value')))
             # TODO: check pam_authz_search expression to only contain
             # username, service, ruser, rhost, tty, hostname, fqdn, dn or
@@ -293,8 +311,21 @@ def read(filename):
             flags = 0 | re.IGNORECASE if m.group('flags') == 'i' else 0
             validnames = re.compile(m.group('value'), flags=flags)
             continue
+        # reconnect_invalidate <MAP>,<MAP>,...
+        m = re.match('reconnect_invalidate\s+(?P<value>\S.*)',
+                     line, re.IGNORECASE)
+        if m:
+            dbs = re.split('[ ,]+', m.group('value').lower())
+            for db in dbs:
+                if db not in maps.keys() + ['nfsidmap']:
+                    raise ParseError(filename, lineno, 'map %s unknown' % db)
+            reconnect_invalidate.update(dbs)
+            continue
         # unrecognised line
         raise ParseError(filename, lineno, 'error parsing line %r' % line)
+    # if logging is not configured, default to syslog
+    if not logs:
+        logs.append(('syslog', logging.INFO))
     # dump config (debugging code)
     for k, v in globals().items():
         if not k.startswith('_'):
