@@ -1,7 +1,7 @@
 
 # ether.py - lookup functions for ethernet addresses
 #
-# Copyright (C) 2010, 2011, 2012 Arthur de Jong
+# Copyright (C) 2010, 2011, 2012, 2013 Arthur de Jong
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -23,6 +23,7 @@ import struct
 import cache
 import common
 import constants
+import search
 
 
 def ether_aton(ether):
@@ -40,15 +41,32 @@ attmap = common.Attributes(cn='cn', macAddress='macAddress')
 filter = '(objectClass=ieee802Device)'
 
 
-class Search(common.Search):
+class Search(search.LDAPSearch):
 
     case_insensitive = ('cn', )
     limit_attributes = ('cn', 'macAddress')
     required = ('cn', 'macAddress')
 
+    def mk_filter(self):
+        # we need a custom mk_filter because this is an | query
+        if 'macAddress' in self.parameters:
+            ether = self.parameters['macAddress']
+            alt_ether = ':'.join('%02x' % int(x, 16) for x in ether.split(':'))
+            return '(&%s(|(%s=%s)(%s=%s)))' % (self.filter,
+                      attmap['macAddress'], ether,
+                      attmap['macAddress'], alt_ether)
+        return super(Search, self).mk_filter()
+
 
 class Cache(cache.Cache):
-    pass
+
+    create_sql = '''
+        CREATE TABLE IF NOT EXISTS `ether_cache`
+          ( `cn` TEXT NOT NULL COLLATE NOCASE,
+            `macAddress` TEXT NOT NULL COLLATE NOCASE,
+            `mtime` TIMESTAMP NOT NULL,
+            UNIQUE (`cn`, `macAddress`) );
+    '''
 
 
 class EtherRequest(common.Request):
