@@ -2,7 +2,7 @@
 
 # pynslcd.py - main daemon module
 #
-# Copyright (C) 2010, 2011, 2012 Arthur de Jong
+# Copyright (C) 2010-2014 Arthur de Jong
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -19,7 +19,6 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 # 02110-1301 USA
 
-import daemon
 import logging
 import logging.handlers
 import os
@@ -28,14 +27,16 @@ import sys
 import syslog
 import threading
 
+import daemon
 import ldap
 
 from tio import TIOStream
-import cfg        # from nslcd.conf
+import cfg
 import common
-import config     # from configure
-import constants  # from nslcd.h
+import constants
+import invalidator
 import mypidfile
+import search
 
 
 # the name of the program
@@ -44,6 +45,9 @@ program_name = 'pynslcd'
 # flag to indicate whether we are in debugging mode
 debugging = 0
 
+# flag to indicate we shouldn't daemonize
+nofork = False
+
 # flag to indicate user requested the --check option
 checkonly = False
 
@@ -51,10 +55,8 @@ checkonly = False
 class MyFormatter(logging.Formatter):
 
     def format(self, record):
-        msg = super(MyFormatter, self).format(record)
-        if record.levelno == logging.DEBUG:
-            msg = 'DEBUG: %s' % msg
-        return msg
+        record.prefix = 'DEBUG: ' if record.levelno == logging.DEBUG else ''
+        return super(MyFormatter, self).format(record)
 
 
 class MySysLogHandler(logging.Handler):
@@ -74,16 +76,13 @@ class MySysLogHandler(logging.Handler):
     def emit(self, record):
         priority = self.mapping.get(record.levelno, syslog.LOG_WARNING)
         msg = self.format(record)
-        for l in msg.splitlines():
-            syslog.syslog(priority, l)
+        for line in msg.splitlines():
+            syslog.syslog(priority, line)
 
 
 # configure logging
-formatter = MyFormatter('%(message)s')
 stderrhandler = logging.StreamHandler(sys.stderr)
-stderrhandler.setFormatter(formatter)
-sysloghandler = MySysLogHandler()
-sysloghandler.setFormatter(formatter)
+stderrhandler.setFormatter(MyFormatter('pynslcd: %(prefix)s%(message)s'))
 logging.getLogger().addHandler(stderrhandler)
 logging.getLogger().setLevel(logging.INFO)
 
@@ -92,10 +91,10 @@ def display_version(fp):
     fp.write('%(PACKAGE_STRING)s\n'
              'Written by Arthur de Jong.\n'
              '\n'
-             'Copyright (C) 2010-2012 Arthur de Jong\n'
+             'Copyright (C) 2010-2014 Arthur de Jong\n'
              'This is free software; see the source for copying conditions.  There is NO\n'
              'warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n'
-             % {'PACKAGE_STRING': config.PACKAGE_STRING, })
+             % {'PACKAGE_STRING': constants.PACKAGE_STRING, })
 
 
 def display_usage(fp):
@@ -103,12 +102,13 @@ def display_usage(fp):
              "Name Service LDAP connection daemon.\n"
              "  -c, --check        check if the daemon already is running\n"
              "  -d, --debug        don't fork and print debugging to stderr\n"
+             "  -n, --nofork       don't fork\n"
              "      --help         display this help and exit\n"
              "      --version      output version information and exit\n"
              "\n"
              "Report bugs to <%(PACKAGE_BUGREPORT)s>.\n"
              % {'program_name': program_name,
-                'PACKAGE_BUGREPORT': config.PACKAGE_BUGREPORT, })
+                'PACKAGE_BUGREPORT': constants.PACKAGE_BUGREPORT, })
 
 
 def parse_cmdline():
@@ -117,8 +117,9 @@ def parse_cmdline():
     global program_name
     program_name = sys.argv[0] or program_name
     try:
-        optlist, args = getopt.gnu_getopt(sys.argv[1:],
-          'cdhV', ('check', 'debug', 'help', 'version', ))
+        optlist, args = getopt.gnu_getopt(
+            sys.argv[1:], 'cdnhV',
+            ('check', 'debug', 'nofork', 'help', 'version'))
         for flag, arg in optlist:
             if flag in ('-c', '--check'):
                 global checkonly
@@ -126,6 +127,9 @@ def parse_cmdline():
             elif flag in ('-d', '--debug'):
                 global debugging
                 debugging += 1
+            elif flag in ('-n', '--nofork'):
+                global nofork
+                nofork = True
             elif flag in ('-h', '--help'):
                 display_usage(sys.stdout)
                 sys.exit(0)
@@ -135,10 +139,12 @@ def parse_cmdline():
         if len(args):
             raise getopt.GetoptError('unrecognized option \'%s\'' % args[0], args[0])
     except getopt.GetoptError, reason:
-        sys.stderr.write("%(program_name)s: %(reason)s\n"
-                         "Try '%(program_name)s --help' for more information.\n"
-                          % {'program_name': program_name,
-                             'reason': reason, })
+        sys.stderr.write(
+            "%(program_name)s: %(reason)s\n"
+            "Try '%(program_name)s --help' for more information.\n" % {
+                'program_name': program_name,
+                'reason': reason,
+            })
         sys.exit(1)
 
 
@@ -149,15 +155,15 @@ def create_socket():
     sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
     # remove existing named socket
     try:
-        os.unlink(config.NSLCD_SOCKET)
+        os.unlink(constants.NSLCD_SOCKET)
     except OSError:
         pass  # ignore any problems
     # bind to named socket
-    sock.bind((config.NSLCD_SOCKET))
+    sock.bind(constants.NSLCD_SOCKET)
     # close the file descriptor on exit
     fcntl.fcntl(sock, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
     # set permissions of socket so anybody can do requests
-    os.chmod(config.NSLCD_SOCKET, 0666)
+    os.chmod(constants.NSLCD_SOCKET, 0666)
     # start listening for connections
     sock.listen(socket.SOMAXCONN)
     return sock
@@ -172,25 +178,30 @@ def getpeercred(fd):
     """Return uid, gid and pid of calling application."""
     import struct
     import socket
-    SO_PEERCRED = 17
-    creds = fd.getsockopt(socket.SOL_SOCKET, SO_PEERCRED, struct.calcsize('3i'))
-    pid, uid, gid = struct.unpack('3i', creds)
-    return uid, gid, pid
+    try:
+        SO_PEERCRED = getattr(socket, 'SO_PEERCRED', 17)
+        creds = fd.getsockopt(socket.SOL_SOCKET, SO_PEERCRED, struct.calcsize('3i'))
+        pid, uid, gid = struct.unpack('3i', creds)
+        return uid, gid, pid
+    except socket.error:
+        return None, None, None
 
 
 handlers = {}
+handlers.update(common.get_handlers('config'))
 handlers.update(common.get_handlers('alias'))
 handlers.update(common.get_handlers('ether'))
 handlers.update(common.get_handlers('group'))
 handlers.update(common.get_handlers('host'))
 handlers.update(common.get_handlers('netgroup'))
 handlers.update(common.get_handlers('network'))
-handlers.update(common.get_handlers('pam'))
 handlers.update(common.get_handlers('passwd'))
 handlers.update(common.get_handlers('protocol'))
 handlers.update(common.get_handlers('rpc'))
 handlers.update(common.get_handlers('service'))
 handlers.update(common.get_handlers('shadow'))
+handlers.update(common.get_handlers('pam'))
+handlers.update(common.get_handlers('usermod'))
 
 
 def acceptconnection(session):
@@ -199,15 +210,11 @@ def acceptconnection(session):
     # See: http://docs.python.org/library/socket.html#socket.socket.settimeout
     fp = None
     try:
-        # probably use finally
-        # indicate new connection to logging module (genrates unique id)
+        # indicate new connection to logging module (generates unique id)
         log_newsession()
         # log connection
-        try:
-            uid, gid, pid = getpeercred(conn)
-            logging.debug('connection from pid=%r uid=%r gid=%r', pid, uid, gid)
-        except:
-            raise  # FIXME: handle exception gracefully
+        uid, gid, pid = getpeercred(conn)
+        logging.debug('connection from pid=%r uid=%r gid=%r', pid, uid, gid)
         # create a stream object
         fp = TIOStream(conn)
         # read request
@@ -219,7 +226,7 @@ def acceptconnection(session):
         try:
             handler = handlers[action]
         except KeyError:
-            logging.warning('invalid action id: %r', action)
+            logging.warning('invalid action id: 0x%08x', action)
             return
         handler(fp, session, uid)()
     finally:
@@ -230,35 +237,23 @@ def acceptconnection(session):
 def disable_nss_ldap():
     """Disable the nss_ldap module to avoid lookup loops."""
     import ctypes
-    lib = ctypes.CDLL(config.NSS_LDAP_SONAME)
-    ctypes.c_int.in_dll(lib, '_nss_ldap_enablelookups').value = 0
-
-
-def get_connection():
-    """Return a connection to the LDAP server."""
-    session = ldap.initialize(cfg.uri)
-    # set session-specific LDAP options
-    if cfg.ldap_version:
-        session.set_option(ldap.OPT_PROTOCOL_VERSION, cfg.ldap_version)
-    if cfg.deref:
-        session.set_option(ldap.OPT_DEREF, cfg.deref)
-    if cfg.timelimit:
-        session.set_option(ldap.OPT_TIMELIMIT, cfg.timelimit)
-        session.set_option(ldap.OPT_TIMEOUT, cfg.timelimit)
-        session.set_option(ldap.OPT_NETWORK_TIMEOUT, cfg.timelimit)
-    if cfg.referrals:
-        session.set_option(ldap.OPT_REFERRALS, cfg.referrals)
-    if cfg.sasl_canonicalize is not None:
-        session.set_option(ldap.OPT_X_SASL_NOCANON, not cfg.sasl_canonicalize)
-    session.set_option(ldap.OPT_RESTART, True)
-    # TODO: register a connection callback (like dis?connect_cb() in myldap.c)
-    if cfg.ssl or cfg.uri.startswith('ldaps://'):
-        session.set_option(ldap.OPT_X_TLS, ldap.OPT_X_TLS_HARD)
-    return session
+    try:
+        lib = ctypes.CDLL(constants.NSS_LDAP_SONAME)
+    except OSError:
+        return  # ignore errors in opening NSS module
+    try:
+        ctypes.c_int.in_dll(lib, '_nss_ldap_enablelookups').value = 0
+    except ValueError:
+        logging.warn('probably older NSS module loaded', exc_info=True)
+    try:
+        version_info = (ctypes.c_char_p * 2).in_dll(lib, '_nss_ldap_version')
+        logging.debug('NSS_LDAP %s %s', version_info[0], version_info[1])
+    except ValueError:
+        logging.warn('probably older NSS module loaded', exc_info=True)
 
 
 def worker():
-    session = get_connection()
+    session = search.Connection()
     while True:
         try:
             acceptconnection(session)
@@ -272,19 +267,19 @@ if __name__ == '__main__':
     parse_cmdline()
     # clean the environment
     os.environ.clear()
-    os.putenv('HOME', '/')
-    os.putenv('TMPDIR', '/tmp')
-    os.putenv('LDAPNOINIT', '1')
-    # disable ldap lookups of host names to avoid lookup loop
-    disable_nss_ldap()
+    os.environ['HOME'] = '/'
+    os.environ['TMPDIR'] = '/tmp'
+    os.environ['LDAPNOINIT'] = '1'
     # set log level
     if debugging:
         logging.getLogger().setLevel(logging.DEBUG)
+    # disable ldap lookups of host names to avoid lookup loop
+    disable_nss_ldap()
     # TODO: implement
     #if myldap_set_debuglevel(cfg.debug) != LDAP_SUCCESS:
     #    sys.exit(1)
     # read configuration file
-    cfg.read(config.NSLCD_CONF_PATH)
+    cfg.read(constants.NSLCD_CONF_PATH)
     # set process title
     try:
         import setproctitle
@@ -294,25 +289,25 @@ if __name__ == '__main__':
     # set a default umask for the pidfile and socket
     os.umask(0022)
     # see if someone already locked the pidfile
-    pidfile = mypidfile.MyPIDLockFile(config.NSLCD_PIDFILE)
+    pidfile = mypidfile.MyPIDLockFile(constants.NSLCD_PIDFILE)
     # see if --check option was given
     if checkonly:
         if pidfile.is_locked():
-            logging.debug('pidfile (%s) is locked', config.NSLCD_PIDFILE)
+            logging.debug('pidfile (%s) is locked', constants.NSLCD_PIDFILE)
             sys.exit(0)
         else:
-            logging.debug('pidfile (%s) is not locked', config.NSLCD_PIDFILE)
+            logging.debug('pidfile (%s) is not locked', constants.NSLCD_PIDFILE)
             sys.exit(1)
     # normal check for pidfile locked
     if pidfile.is_locked():
         logging.error('daemon may already be active, cannot acquire lock (%s)',
-                      config.NSLCD_PIDFILE)
+                      constants.NSLCD_PIDFILE)
         sys.exit(1)
     # daemonize
-    if debugging:
-        daemon = pidfile
+    if debugging or nofork:
+        ctx = pidfile
     else:
-        daemon = daemon.DaemonContext(
+        ctx = daemon.DaemonContext(
                       pidfile=pidfile,
                       signal_map={
                           signal.SIGTERM: 'terminate',
@@ -320,12 +315,25 @@ if __name__ == '__main__':
                           signal.SIGPIPE: None,
                       })
     # start daemon
-    with daemon:
-        # start normal logging to syslog
-        if not debugging:
-            logging.getLogger().addHandler(sysloghandler)
-        logging.info('version %s starting', config.VERSION)
+    with ctx:
         try:
+            # start normal logging as configured
+            if not debugging:
+                for method, level in cfg.logs:
+                    if method == 'syslog':
+                        handler = MySysLogHandler()
+                        handler.setFormatter(MyFormatter('%(prefix)s%(message)s'))
+                    else:
+                        handler = logging.FileHandler(method, encoding='utf-8')
+                        handler.setFormatter(MyFormatter('%(asctime)s %(prefix)s%(message)s'))
+                    handler.setLevel(level)
+                    logging.getLogger().addHandler(handler)
+                logging.getLogger().setLevel(min(level for method, level in cfg.logs))
+                logging.getLogger().removeHandler(stderrhandler)
+            logging.info('version %s starting', constants.VERSION)
+            # start invalidator sub-process if needed
+            if cfg.reconnect_invalidate:
+                invalidator.start_invalidator()
             # create socket
             nslcd_serversocket = create_socket()
             # load supplementary groups
